#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import os
import pathlib
import shutil
from operator import itemgetter

from girder import events
from girder.api import access
from girder.constants import AccessType, TokenScope
from girder.exceptions import (
    AccessException,
    ResourcePathNotFound,
    RestException,
    ValidationException,
)
from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.user import User
from girder.utility.model_importer import ModelImporter
from girder.utility.path import getResourcePath, lookUpToken, split
from girder.utility.progress import ProgressContext

from . import VirtualObject, ensure_unique_path, validate_event


class EmptyDocumentError(Exception):
    pass


class VirtualResource(VirtualObject):
    def __init__(self):
        super(VirtualResource, self).__init__()
        self.resourceName = "virtual_resource"
        name = "virtual_resources"

        events.bind("rest.delete.resource.before", name, self.delete_resources)
        # GET /resource/:id
        events.bind("rest.get.resource/:id/path.before", name, self.path)
        # PUT /resource/:id/timestamp
        events.bind("rest.post.resource/copy.before", name, self.copy_resources)
        # GET /resource/:id/download
        events.bind("rest.get.resource/lookup.before", name, self.lookup)
        events.bind("rest.put.resource/move.before", name, self.move_resources)
        # GET /resource/search

    def _filter_resources(self, event, level=AccessType.WRITE, user=None):
        resources = json.loads(event.info["params"]["resources"])
        remaining_resources = {"folder": [], "item": []}
        wt_resources = []
        for kind in resources:
            for obj_id in resources[kind]:
                if obj_id.startswith("wtlocal:"):
                    source_path, root_id = self.path_from_id(obj_id)
                    try:
                        root = Folder().load(root_id, user=user, level=level, exc=False)
                    except AccessException:
                        root = None
                    if root:
                        wt_resources.append({"src_path": source_path, "kind": kind})
                else:
                    remaining_resources[kind].append(obj_id)
        total = sum([len(remaining_resources[key]) for key in remaining_resources])
        event.info["params"]["resources"] = json.dumps(remaining_resources)
        if total == 0:
            event.preventDefault().addResponse(None)
        return sorted(
            wt_resources, key=itemgetter("kind"), reverse=True
        )  # We want to have items first, which is relevant for MOVE op

    @access.user(scope=TokenScope.DATA_OWN)
    def delete_resources(self, event):
        user = self.getCurrentUser()
        wt_resources = self._filter_resources(event, level=AccessType.WRITE, user=user)
        progress = event.info["params"].get("progress", False)
        with ProgressContext(
            progress,
            user=user,
            title="Deleting resources",
            message="Calculating requirements...",
            total=len(wt_resources),
        ) as ctx:
            for obj in wt_resources:
                source_path = obj["src_path"]
                ctx.update(message="Deleting %s %s" % (obj["kind"], source_path.name))
                if obj["kind"] == "folder":
                    shutil.rmtree(source_path.as_posix())
                else:
                    source_path.unlink()
                ctx.update(increment=1)

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def copy_resources(self, event, path, root, user=None):
        wt_resources = self._filter_resources(event, level=AccessType.READ, user=user)
        progress = event.info["params"].get("progress", False)
        with ProgressContext(
            progress,
            user=user,
            title="Copying resources",
            message="Calculating requirements...",
            total=len(wt_resources),
        ) as ctx:
            for obj in wt_resources:
                source_path = obj["src_path"]
                ctx.update(message="Copying %s %s" % (obj["kind"], source_path.name))
                if obj["kind"] == "folder":
                    shutil.copytree(
                        source_path.as_posix(), (path / source_path.name).as_posix()
                    )
                else:
                    name = source_path.name
                    new_path = ensure_unique_path(path, name)
                    shutil.copy(source_path.as_posix(), new_path.as_posix())
                ctx.update(increment=1)

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def move_resources(self, event, path, root, user=None):
        wt_resources = self._filter_resources(event, level=AccessType.WRITE, user=user)
        progress = event.info["params"].get("progress", False)
        with ProgressContext(
            progress,
            user=user,
            title="Moving resources",
            message="Calculating requirements...",
            total=len(wt_resources),
        ) as ctx:
            for obj in wt_resources:
                source_path = obj["src_path"]
                ctx.update(message="Moving %s %s" % (obj["kind"], source_path.name))
                shutil.move(
                    source_path.as_posix(), (path / source_path.name).as_posix()
                )
                ctx.update(increment=1)

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def path(self, event, path, root, user=None):
        res_type = event.info["params"]["type"]
        try:
            if res_type == "folder":
                self.is_dir(path, root)
            elif res_type in ("item", "file"):
                self.is_file(path, root)
            else:
                raise ValidationException("Not a file, nor a folder")
        except ValidationException:
            raise RestException("Invalid resource id.")

        root_girder_path = pathlib.Path(getResourcePath("folder", root, user=user))
        remainder_path = path.relative_to(pathlib.PosixPath(root["fsPath"]))
        response = (root_girder_path / remainder_path).as_posix()
        event.preventDefault().addResponse(response)

    @access.public(scope=TokenScope.DATA_READ)
    def lookup(self, event):
        test = event.info["params"].get("test", False)
        path = event.info["params"].get("path")
        if len(split(path.lstrip("/"))) < 3:
            return  # either a user or a collection, so default to core
        response = self._lookUpPath(path, self.getCurrentUser(), test)["document"]
        event.preventDefault().addResponse(response)

    @staticmethod
    def _lookup_err(msg, test=False):
        if test:
            raise EmptyDocumentError
        else:
            raise ResourcePathNotFound(msg)

    def _get_base(self, pathArray, test=False):
        model = pathArray[0]
        if model == "user":
            username = pathArray[1]
            parent = User().findOne({"login": username})
            if parent is None:
                self._lookup_err("User not found: %s" % username, test=test)
        elif model == "collection":
            collectionName = pathArray[1]
            parent = Collection().findOne({"name": collectionName})
            if parent is None:
                self._lookup_err("Collection not found: %s" % collectionName, test=test)
        else:
            raise ValidationException("Invalid path format")
        return parent, model

    def _get_vobject(self, document, path, i):
        pathArray = split(path)
        root = document
        n = 3 + i
        fspath = os.path.join(document["fsPath"], "/".join(pathArray[n:]))
        fspath = pathlib.Path(fspath)
        if not fspath.exists():
            raise ValidationException("Path not found: %s" % path)
        if fspath.is_dir():
            document = self.vFolder(fspath, root)
            model = "folder"
        elif fspath.is_file():
            document = self.vItem(fspath, root)
            model = "item"
        # TODO: add vLink here...
        return document, model

    def _lookUpPath(self, path, user=None, test=False, filter=True, force=False):
        """
        Look up a resource in the data hierarchy by path.

        :param path: path of the resource
        :param user: user with correct privileges to access path
        :param test: defaults to false, when set to true
            will return None instead of throwing exception when
            path doesn't exist
        :type test: bool
        :param filter: Whether the returned model should be filtered.
        :type filter: bool
        :param force: if True, don't validate the access.
        :type force: bool
        """
        path = path.lstrip("/")
        pathArray = split(path)

        try:
            document, model = self._get_base(pathArray, test=test)
        except EmptyDocumentError:
            return {"model": None, "document": None}

        try:
            if not force:
                ModelImporter.model(model).requireAccess(document, user)
            token = None
            for i, token in enumerate(pathArray[2:]):  # noqa
                document, model = lookUpToken(token, model, document)
                if (
                    model == "folder"
                    and document.get("isSymlink")
                    and document.get("symlinkTargetId")
                ):
                    document = Folder().load(
                        document["symlinkTargetId"], force=True, exc=True
                    )
                if not force:
                    ModelImporter.model(model).requireAccess(document, user)
                if "fsPath" in document:
                    break
            if token != pathArray[-1]:
                document, model = self._get_vobject(document, path, i)
        except (ValidationException, AccessException):
            if test:
                return {"model": None, "document": None}
            raise ResourcePathNotFound("Path not found: %s" % path)

        if filter:
            document = ModelImporter.model(model).filter(document, user)

        return {"model": model, "document": document}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import errno
import os
import pathlib
import shutil
import stat

import cherrypy

from girder import events
from girder.api import access
from girder.api.rest import setResponseHeader
from girder.constants import AccessType, TokenScope
from girder.exceptions import AccessException, GirderException, RestException
from girder.models.assetstore import Assetstore
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.upload import Upload
from girder.utility import RequestBodyStream, assetstore_utilities

from . import VirtualObject, bail_if_exists, validate_event

BUF_SIZE = 65536
DEFAULT_PERMS = stat.S_IRUSR | stat.S_IWUSR


class VirtualFile(VirtualObject):
    def __init__(self):
        super(VirtualFile, self).__init__()
        self.resourceName = "virtual_file"
        name = "virtual_resources"

        events.bind("rest.post.file.before", name, self.create_file)
        events.bind("rest.get.file/:id.before", name, self.get_file_info)
        events.bind("rest.put.file/:id.before", name, self.rename_file)
        events.bind("rest.delete.file/:id.before", name, self.remove_file)
        # PUT /file/:id/contents
        # POST /file/:id/copy
        events.bind("rest.get.item/:id/download.before", name, self.file_download)
        events.bind("rest.get.file/:id/download.before", name, self.file_download)
        # GET /file/:id/download/:name
        # PUT /file/:id/move
        events.bind("rest.post.file/chunk.before", name, self.read_chunk)
        # POST /file/completion
        # GET /file/offset
        # DELETE /file/upload/:id

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def create_file(self, event, path, root, user=None):
        params = event.info["params"]
        name = params["name"]
        file_path = path / name
        try:
            path.mkdir(parents=True, exist_ok=True)
            with file_path.open(mode="a"):
                os.utime(file_path.as_posix())
        except PermissionError:
            raise GirderException(
                "Insufficient perms to write on {}".format(path.as_posix()),
                "girder.api.v1.file.create-upload-failed",
            )
        except Exception:
            raise

        parent = Folder().filter(self.vFolder(path, root), user=user)
        size = int(params["size"])
        chunk = None
        if size > 0 and cherrypy.request.headers.get("Content-Length"):
            ct = cherrypy.request.body.content_type.value
            if (
                ct not in cherrypy.request.body.processors
                and ct.split("/", 1)[0] not in cherrypy.request.body.processors
            ):
                chunk = RequestBodyStream(cherrypy.request.body)
        if chunk is not None and chunk.getSize() <= 0:
            chunk = None
        upload = Upload().createUpload(
            user=user, name=name, parentType="folder", parent=parent, size=size
        )

        if upload["size"] > 0:
            if chunk:
                fobj = self._handle_chunk(upload, chunk, filter=True, user=user)
                event.preventDefault().addResponse(fobj)
                return
            event.preventDefault().addResponse(upload)
        else:
            event.preventDefault().addResponse(
                File().filter(self.vFile(file_path, root), user=user)
            )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_file_info(self, event, path, root, user=None):
        event.preventDefault().addResponse(
            File().filter(self.vFile(path, root), user=user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def rename_file(self, event, path, root, user=None):
        self.is_file(path, root["_id"])
        new_path = path.with_name(event.info["params"]["name"])
        bail_if_exists(new_path)
        path.rename(new_path)
        event.preventDefault().addResponse(
            File().filter(self.vFile(new_path, root), user=user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def remove_file(self, event, path, root, user=None):
        self.is_file(path, root["_id"])
        path.unlink()
        event.preventDefault().addResponse({"message": "Deleted file %s." % path.name})

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @validate_event(level=AccessType.READ)
    def file_download(self, event, path, root, user=None):
        fobj = self.vFile(path, root)

        rangeHeader = cherrypy.lib.httputil.get_ranges(
            cherrypy.request.headers.get("Range"), fobj.get("size", 0)
        )
        # The HTTP Range header takes precedence over query params
        if rangeHeader and len(rangeHeader):
            # Currently we only support a single range.
            offset, endByte = rangeHeader[0]
        else:
            endByte = min(
                int(event.info["params"].get("endByte", fobj["size"])), fobj["size"]
            )
            offset = int(event.info["params"].get("offset", "0"))

        setResponseHeader("Accept-Ranges", "bytes")
        setResponseHeader("Content-Type", "application/octet-stream")
        setResponseHeader("Content-Length", max(endByte - offset, 0))
        if (offset or endByte < fobj["size"]) and fobj["size"]:
            setResponseHeader(
                "Content-Range", "bytes %d-%d/%d" % (offset, endByte - 1, fobj["size"])
            )
        disp = event.info["params"].get("contentDisposition", "attachment")
        setResponseHeader(
            "Content-Disposition", '{}; filename="{}"'.format(disp, fobj["name"])
        )

        def stream():
            bytesRead = offset
            with path.open(mode="rb") as f:
                if offset > 0:
                    f.seek(offset)

                while True:
                    readLen = min(BUF_SIZE, endByte - bytesRead)
                    if readLen <= 0:
                        break

                    data = f.read(readLen)
                    bytesRead += readLen

                    # if not data:
                    #    break
                    yield data

        event.preventDefault().addResponse(stream)

    def _finalize_upload(self, upload, assetstore=None):
        if assetstore is None:
            assetstore = Assetstore().load(upload["assetstoreId"])
        if str(upload["parentId"]).startswith("wtlocal:"):
            path, root_id = self.path_from_id(upload["parentId"])
            root = Folder().load(root_id, force=True)  # TODO make it obsolete
        else:
            root = Folder().load(upload["parentId"], force=True)
            path = pathlib.Path(root["fsPath"])
        abspath = path / upload["name"]
        shutil.move(upload["tempFile"], abspath.as_posix())
        abspath.chmod(assetstore.get("perms", DEFAULT_PERMS))
        return self.vFile(abspath, root)

    def _handle_chunk(
        self, upload, chunk, filter=False, user=None, uploadExtraParameters=None
    ):
        assetstore = Assetstore().load(upload["assetstoreId"])
        adapter = assetstore_utilities.getAssetstoreAdapter(assetstore)

        upload = adapter.uploadChunk(upload, chunk, uploadExtraParameters)
        if "_id" in upload or upload["received"] != upload["size"]:
            upload = Upload().save(upload)

        # If upload is finished, we finalize it
        if upload["received"] == upload["size"]:
            return self._finalize_upload(upload)
        else:
            return upload

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def read_chunk(self, event, path, root, user=None):
        params = event.info["params"]
        if "chunk" in params:
            chunk = params["chunk"]
            if isinstance(chunk, cherrypy._cpreqbody.Part):
                # Seek is the only obvious way to get the length of the part
                chunk.file.seek(0, os.SEEK_END)
                size = chunk.file.tell()
                chunk.file.seek(0, os.SEEK_SET)
                chunk = RequestBodyStream(chunk.file, size=size)
        else:
            chunk = RequestBodyStream(cherrypy.request.body)

        if not user:
            user = self.getCurrentUser()
        offset = int(params.get("offset", 0))
        upload = Upload().load(params["uploadId"])
        uploadExtraParameters = params.get("uploadExtraParameters")

        if upload["userId"] != user["_id"]:
            raise AccessException("You did not initiate this upload.")

        if upload["received"] != offset:
            raise RestException(
                "Server has received %s bytes, but client sent offset %s."
                % (upload["received"], offset)
            )

        try:
            fobj = self._handle_chunk(
                upload,
                chunk,
                filter=True,
                user=user,
                uploadExtraParameters=uploadExtraParameters,
            )
            event.preventDefault().addResponse(fobj)
        except IOError as exc:
            if exc.errno == errno.EACCES:
                raise Exception("Failed to store upload.")
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import base64
import copy
import datetime
import os
import pathlib

from girder.api.rest import Resource
from girder.constants import AccessType
from girder.exceptions import ValidationException
from girder.models.folder import Folder
from girder.models.upload import Upload


def bail_if_exists(path):
    if path.exists():
        raise ValidationException(
            "A folder or file with that name already exists here.", "name"
        )


def ensure_unique_path(dirname, name):
    checkName = (dirname / name).exists()
    new_name = name
    n = 0
    while checkName:
        n += 1
        new_name = "%s (%d)" % (name, n)
        checkName = (dirname / new_name).exists()
    return dirname / new_name


def validate_event(level=AccessType.READ, validate_admin=False):
    def validation(func):
        def wrapper(self, event):
            params = event.info.get("params", {})
            if {"isMapping", "fsPath"} & set(params) and validate_admin:
                folder = Folder().load(event.info["id"], force=True)
                update = False

                if params.get("isMapping") is not None:
                    update = True
                    folder["isMapping"] = bool(params["isMapping"])
                if params.get("fsPath") is not None:
                    update = True
                    folder["fsPath"] = params["fsPath"]

                if update:
                    self.requireAdmin(
                        self.getCurrentUser(), "Must be admin to setup virtual folders."
                    )
                    folder = Folder().filter(Folder().save(folder), self.getCurrentUser())
                    pathlib.Path(folder["fsPath"]).mkdir(
                        mode=0o755, parents=True, exist_ok=True
                    )
                    event.preventDefault().addResponse(folder)
                    return

            if "uploadId" in params:
                upload = Upload().load(params["uploadId"])
                try:
                    parent_id = str(upload["parentId"])
                    parent_type = upload["parentType"]
                except KeyError:
                    parent_id = str(upload["fileId"])
                    parent_type = None
            else:
                parent_id = params.get("parentId")
                parent_type = params.get("parentType") or "folder"

            obj_id = event.info.get("id")
            folder_id = params.get("folderId")
            item_id = params.get("itemId")
            any_parent_id = parent_id or folder_id or item_id

            path = None
            if obj_id:
                path, root_id = VirtualObject.path_from_id(obj_id)
            elif any_parent_id and any_parent_id.startswith("wtlocal:"):
                path, root_id = VirtualObject.path_from_id(any_parent_id)
            elif (parent_id and parent_type == "folder") or folder_id:
                path, root_id = VirtualObject.path_from_id(parent_id or folder_id)

            if path:
                path = pathlib.Path(path)
                if path.is_absolute():
                    user = self.getCurrentUser()
                    root = Folder().load(root_id, level=level, user=user, exc=True)
                    func(self, event, path, root, user=user)

        return wrapper

    return validation


class VirtualObject(Resource):
    def __init__(self):
        super(VirtualObject, self).__init__()

    @staticmethod
    def generate_id(path, root_id):
        if isinstance(path, pathlib.Path):
            path = path.as_posix()
        path += "|" + str(root_id)
        return "wtlocal:" + base64.b64encode(path.encode()).decode()

    @staticmethod
    def path_from_id(object_id):
        if str(object_id).startswith("wtlocal:"):
            decoded = base64.b64decode(object_id[8:]).decode()
            path, root_id = decoded.split("|")
        else:
            root_folder = Folder().load(object_id, force=True) or {}
            path = root_folder.get("fsPath")  # only exists on virtual folders
            root_id = str(root_folder.get("_id"))
        if path:
            path = pathlib.Path(path)
        return path, root_id

    def is_file(self, path, root_id):
        if not path.is_file():
            raise ValidationException(
                "Invalid ObjectId: %s" % self.generate_id(path, root_id), field="id"
            )

    def is_symlink(self, path, root_id):
        if not path.is_symlink():
            raise ValidationException(
                "Invalid ObjectId: %s" % self.generate_id(path, root_id), field="id"
            )

    def is_dir(self, path, root_id):
        if not path.is_dir():
            raise ValidationException(
                "Invalid ObjectId: %s" % self.generate_id(path, root_id), field="id"
            )

    def vFolder(self, path, root):
        self.is_dir(path, root["_id"])
        stat = path.stat()

        if path == pathlib.Path(root["fsPath"]):
            # We want actual mtime/ctime from disk
            root.update({
                "created": datetime.datetime.fromtimestamp(stat.st_ctime, datetime.timezone.utc),
                "updated": datetime.datetime.fromtimestamp(stat.st_mtime, datetime.timezone.utc),
            })
            return root

        if path.parent == pathlib.Path(root["fsPath"]):
            parentId = root["_id"]
        else:
            parentId = self.generate_id(path.parent.as_posix(), root["_id"])

        return {
            "_id": self.generate_id(path.as_posix(), root["_id"]),
            "_modelType": "folder",
            "access": copy.deepcopy(root.get("access", {"users": [], "groups": []})),
            "name": path.parts[-1],
            "parentId": parentId,
            "parentCollection": "folder",
            "creatorId": None,
            "created": datetime.datetime.fromtimestamp(stat.st_ctime, datetime.timezone.utc),
            "updated": datetime.datetime.fromtimestamp(stat.st_mtime, datetime.timezone.utc),
            "size": stat.st_size,
            "public": root.get("public", False),
            "lowerName": path.parts[-1].lower(),
        }

    def vItem(self, path, root):
        self.is_file(path, root["_id"])
        stat = path.stat()
        return {
            "_id": self.generate_id(path.as_posix(), root["_id"]),
            "_modelType": "item",
            "name": path.parts[-1],
            "folderId": self.generate_id(path.parent.as_posix(), root["_id"]),
            "creatorId": None,
            "created": datetime.datetime.fromtimestamp(stat.st_ctime, datetime.timezone.utc),
            "updated": datetime.datetime.fromtimestamp(stat.st_mtime, datetime.timezone.utc),
            "size": stat.st_size,
            "lowerName": path.parts[-1].lower(),
        }

    def vLink(self, path, root):
        self.is_symlink(path, root["_id"])
        stat = path.lstat()
        return {
            "_id": self.generate_id(path.as_posix(), root["_id"]),
            "_modelType": None,
            "name": path.parts[-1],
            "folderId": self.generate_id(path.parent.as_posix(), root["_id"]),
            "creatorId": None,
            "created": datetime.datetime.fromtimestamp(stat.st_ctime, datetime.timezone.utc),
            "updated": datetime.datetime.fromtimestamp(stat.st_mtime, datetime.timezone.utc),
            "size": stat.st_size,
            "lowerName": path.parts[-1].lower(),
            "linkTarget": os.readlink(path)
        }

    def vFile(self, path, root):
        self.is_file(path, root["_id"])
        stat = path.stat()
        return {
            "_id": self.generate_id(path.as_posix(), root["_id"]),
            "_modelType": "file",
            "assetstoreId": None,
            "creatorId": root["creatorId"],
            "mimeType": "application/octet-stream",
            "name": path.parts[-1],
            "size": stat.st_size,
            "exts": [_[1:] for _ in path.suffixes],
            "created": datetime.datetime.fromtimestamp(stat.st_ctime, datetime.timezone.utc),
            "updated": datetime.datetime.fromtimestamp(stat.st_mtime, datetime.timezone.utc),
            "itemId": self.generate_id(path.as_posix(), root["_id"]),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path

from girder.constants import AccessType
from girder.models.folder import Folder
from girder.plugin import GirderPlugin, registerPluginStaticContent

from .rest.virtual_file import VirtualFile
from .rest.virtual_folder import VirtualFolder
from .rest.virtual_item import VirtualItem
from .rest.virtual_resource import VirtualResource


class VirtualResourcesPlugin(GirderPlugin):
    DISPLAY_NAME = "Virtual Resources"

    def load(self, info):
        from girder.api.v1.folder import (
            Folder as FolderResource,  # noqa: F401 circular import
        )

        Folder().exposeFields(level=AccessType.READ, fields={"isMapping", "isSymlink"})
        Folder().exposeFields(
            level=AccessType.SITE_ADMIN, fields={"fsPath", "symlinkTargetId"}
        )
        for endpoint in (FolderResource.updateFolder, FolderResource.createFolder):
            (
                endpoint.description.param(
                    "isMapping",
                    "Whether this is a virtual folder.",
                    required=False,
                    dataType="boolean",
                )
                .param("fsPath", "Local filesystem path it maps to.", required=False)
                .param(
                    "isSymlink",
                    "Set to true to create a symlink to another folder.",
                    required=False,
                    dataType="boolean",
                )
                .param(
                    "symlinkTargetId",
                    "The ID of the target folder to which this folder will be a symlink.",
                    required=False,
                    dataType="string",
                )
            )

        info["apiRoot"].virtual_item = VirtualItem()
        virtual_file = VirtualFile()
        info["apiRoot"].virtual_file = virtual_file
        info["apiRoot"].virtual_folder = VirtualFolder()
        info["apiRoot"].virtual_resource = VirtualResource()

        registerPluginStaticContent(
            plugin="virtual_resources",
            css=["/style.css"],
            js=["/girder-plugin-virtual-resources.umd.cjs"],
            staticDir=Path(__file__).parent / "web_client" / "dist",
            tree=info["serverRoot"],
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import pathlib
import shutil
from operator import itemgetter

from girder import events
from girder.api import access
from girder.constants import AccessType, TokenScope
from girder.exceptions import GirderException, ValidationException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item

import pymongo

from . import VirtualObject, bail_if_exists, ensure_unique_path, validate_event


class VirtualItem(VirtualObject):
    def __init__(self):
        super(VirtualItem, self).__init__()
        self.resourceName = "virtual_item"
        name = "virtual_resources"

        events.bind("model.item.validate", name, self.validate_item)
        events.bind("rest.get.item.before", name, self.get_child_items)
        events.bind("rest.post.item.before", name, self.create_item)
        events.bind("rest.get.item/:id.before", name, self.get_item_info)
        events.bind("rest.put.item/:id.before", name, self.rename_item)
        events.bind("rest.delete.item/:id.before", name, self.remove_item)
        events.bind("rest.post.item/:id/copy.before", name, self.copy_item)
        # events.bind("rest.get.item/:id/download.before", name, self.file_download)  # in Vfile
        events.bind("rest.get.item/:id/files.before", name, self.get_child_files)
        # PUT/DELETE /item/:id/metadata
        events.bind("rest.get.item/:id/rootpath.before", name, self.item_root_path)

    @staticmethod
    def validate_item(event):
        doc = event.info
        if isinstance(doc.get("_id"), str) and doc.get("_id").startswith("wtlocal:"):
            return
        parent = Folder().load(doc["folderId"], force=True, exc=True)
        if parent.get("isSymlink"):
            raise ValidationException(
                "You may not place items under a symlink folder.", field="folderId"
            )

    @access.public(scope=TokenScope.DATA_READ)
    def get_child_items(self, event):
        params = event.info["params"]
        folder_id = params.get("folderId")
        if folder_id and not folder_id.startswith("wtlocal:"):
            parent = Folder().load(params["folderId"], force=True, exc=True)
            if parent.get("isSymlink") and parent.get("symlinkTargetId"):
                event.info["params"]["folderId"] = str(parent["symlinkTargetId"])
        self._get_child_items(event)

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def _get_child_items(self, event, path, root, user=None):
        params = event.info["params"]

        self.is_dir(path, root["_id"])
        offset = int(params.get("offset", 0))
        limit = int(params.get("limit", 50))
        name = params.get("name")
        sort_key = params.get("sort", "lowerName")
        reverse = int(params.get("sortdir", pymongo.ASCENDING)) == pymongo.DESCENDING

        if name:
            if (path / name).is_file():
                items = [self.vItem(path / name, root)]
            else:
                items = []
        else:
            items = [self.vItem(obj, root) for obj in path.iterdir() if obj.is_file()]
        items = sorted(items, key=itemgetter(sort_key), reverse=reverse)
        upper_bound = limit + offset if limit > 0 else None
        response = [
            Item().filter(item, user=user) for item in items[offset:upper_bound]
        ]
        event.preventDefault().addResponse(response)

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def create_item(self, event, path, root, user=None):
        params = event.info["params"]
        new_path = path / params["name"]
        try:
            path.mkdir(parents=True, exist_ok=True)
            with new_path.open(mode="x"):
                os.utime(new_path.as_posix())
        except FileExistsError:
            raise ValidationException(
                "An item with that name already exists here.", "name"
            )
        event.preventDefault().addResponse(
            Item().filter(self.vItem(new_path, root), user=user)
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_item_info(self, event, path, root, user=None):
        event.preventDefault().addResponse(
            Item().filter(self.vItem(path, root), user=user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def rename_item(self, event, path, root, user=None):
        self.is_file(path, root["_id"])
        source = self.vItem(path, root)

        params = event.info.get("params", {})
        name = params.get("name", path.name)
        parentId = params.get("folderId", source["folderId"])

        if parentId == source["folderId"]:
            new_path = path.with_name(name)
            bail_if_exists(new_path)
            path.rename(new_path)
        else:
            dst_path, dst_root_id = self.path_from_id(parentId)
            if not dst_path:
                raise GirderException("Folder {} is not a mapping.".format(parentId))
            # Check whether the user can write to the destination
            Folder().load(dst_root_id, user=user, level=AccessType.WRITE, exc=True)
            self.is_dir(dst_path, dst_root_id)
            new_path = dst_path / name
            bail_if_exists(new_path)
            shutil.move(path.as_posix(), new_path.as_posix())

        event.preventDefault().addResponse(
            Item().filter(self.vItem(new_path, root), user=user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def remove_item(self, event, path, root, user=None):
        self.is_file(path, root["_id"])
        path.unlink()
        event.preventDefault().addResponse({"message": "Deleted item %s." % path.name})

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def copy_item(self, event, path, root, user=None):
        self.is_file(path, root["_id"])
        source = self.vItem(path, root)
        name = event.info["params"].get("name") or path.name

        folder_id = event.info["params"].get("folderId", source["folderId"])
        if str(folder_id).startswith("wtlocal:"):
            new_dirname, new_root_id = self.path_from_id(folder_id)
        else:
            new_root = Folder().load(folder_id, force=True, exc=True)
            try:
                new_dirname = pathlib.Path(new_root["fsPath"])
            except KeyError:
                raise GirderException("Folder {} is not a mapping.".format(folder_id))
            new_root_id = str(new_root["_id"])

        if new_root_id != str(root["_id"]):
            new_root = Folder().load(
                new_root["_id"], user=user, level=AccessType.WRITE, exc=True
            )  # Check ACLs
        else:
            new_root = root

        new_path = ensure_unique_path(new_dirname, name)
        shutil.copy(path.as_posix(), new_path.as_posix())
        event.preventDefault().addResponse(
            Item().filter(self.vItem(new_path, new_root), user=user)
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_child_files(self, event, path, root, user=None):
        event.preventDefault().addResponse(
            [File().filter(self.vFile(path, root), user=user)]
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def item_root_path(self, event, path, root, user=None):
        root_path = pathlib.Path(root["fsPath"])
        response = [
            {"type": "item", "object": Item().filter(self.vItem(path, root), user=user)}
        ]
        path = path.parent
        while path != root_path:
            response.append(
                {
                    "type": "folder",
                    "object": Folder().filter(self.vFolder(path, root), user=user),
                }
            )
            path = path.parent

        response.append({"type": "folder", "object": Folder().filter(root, user=user)})
        girder_rootpath = Folder().parentsToRoot(root, user=user)
        response += girder_rootpath[::-1]
        response.pop(0)
        event.preventDefault().addResponse(response[::-1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import pathlib
import shutil
from operator import itemgetter

import pymongo
from bson.objectid import ObjectId
from girder import events
from girder.api import access
from girder.api.rest import setContentDisposition, setResponseHeader
from girder.constants import AccessType, TokenScope
from girder.exceptions import GirderException, ValidationException
from girder.models.folder import Folder
from girder.utility import ziputil

from . import VirtualObject, bail_if_exists, ensure_unique_path, validate_event


def file_stream(path, buf_size=65536):
    bytes_read = 0
    end_byte = path.stat().st_size
    with path.open(mode="rb") as f:
        while True:
            read_len = min(buf_size, end_byte - bytes_read)
            if read_len <= 0:
                break
            data = f.read(read_len)
            bytes_read += read_len
            # if not data:
            #    break
            yield data


class VirtualFolder(VirtualObject):
    def __init__(self):
        super(VirtualFolder, self).__init__()
        self.resourceName = "virtual_folder"
        name = "virtual_resources"
        events.bind("model.folder.validate", "jsonforms", self._validateFolder)
        events.bind("rest.get.folder.before", name, self.get_child_folders)
        events.bind("rest.post.folder.before", name, self.create_folder)
        events.bind("rest.get.folder/:id.before", name, self.get_folder_info)
        events.bind("rest.put.folder/:id.before", name, self.rename_folder)
        events.bind("rest.delete.folder/:id.before", name, self.remove_folder)
        events.bind("rest.post.folder.after", "jsonforms", self._folderUpdate)
        events.bind("rest.put.folder/:id.after", "jsonforms", self._folderUpdate)
        # GET /folder/:id/access -- not needed
        # PUT /folder/:id/access -- not needed
        # PUT /folder/:id/check -- not needed
        events.bind(
            "rest.delete.folder/:id/contents.before", name, self.remove_folder_contents
        )
        events.bind("rest.post.folder/:id/copy.before", name, self.copy_folder)
        events.bind("rest.get.folder/:id/details.before", name, self.get_folder_details)
        events.bind("rest.get.folder/:id/download.before", name, self.download_folder)
        # PUT/DELETE /folder/:id/metadata -- not needed
        events.bind("rest.get.folder/:id/rootpath.before", name, self.folder_root_path)
        events.bind(
            "rest.post.folder/recursive.before", name, self.create_folder_recursive
        )
        # For README plugin
        events.bind("rest.get.folder/:id/readme.before", name, self.get_folder_readme)

    def _folderUpdate(self, event):
        params = event.info["params"]
        if {"isSymlink", "symlinkTargetId"} & set(params):
            folder = Folder().load(event.info["returnVal"]["_id"], force=True)
            update = False

            if params.get("isSymlink") is not None:
                update = True
                folder["isSymlink"] = params["isSymlink"]
            if params.get("symlinkTargetId"):
                update = True
                try:
                    folder["symlinkTargetId"] = ObjectId(params["symlinkTargetId"])
                except Exception:
                    raise ValidationException(
                        "symlinkTargetId must be an ObjectId", field="symlinkTargetId"
                    )

            if update and not folder.get("isSymlink"):
                folder["symlinkTargetId"] = None

            if update:
                self.requireAdmin(
                    self.getCurrentUser(), "Must be admin to setup symlink folders."
                )
                folder = Folder().filter(Folder().save(folder), self.getCurrentUser())
                event.preventDefault().addResponse(folder)

    @staticmethod
    def _validateFolder(event):
        doc = event.info

        if isinstance(doc.get("_id"), str) and doc.get("_id").startswith("wtlocal:"):
            return

        if "isSymlink" in doc and not isinstance(doc["isSymlink"], bool):
            raise ValidationException("isSymlink must be a boolean.", field="isSymlink")

        if doc.get("isSymlink"):
            # Make sure it doesn't have children
            if list(Folder().childItems(doc, limit=1)):
                raise ValidationException(
                    "Symlink folders may not contain child items.", field="isSymlink"
                )
            if list(
                Folder().find(
                    {"parentId": doc["_id"], "parentCollection": "folder"}, limit=1
                )
            ):
                raise ValidationException(
                    "Symlink folders may not contain child folders.", field="isSymlink"
                )
        if doc["parentCollection"] == "folder":
            parent = Folder().load(event.info["parentId"], force=True, exc=True)
            if parent.get("isSymlink"):
                raise ValidationException(
                    "You may not place folders under a symlink folder.",
                    field="folderId",
                )

        if doc.get("symlinkTargetId") and not isinstance(
            doc["symlinkTargetId"], ObjectId
        ):
            raise ValidationException(
                "symlinkTargetId must be an ObjectId", field="symlinkTargetId"
            )

        if doc.get("symlinkTargetId"):
            if doc["_id"] == doc["symlinkTargetId"]:
                raise ValidationException(
                    "A folder may not symlink to itself.", field="symlinkTargetId"
                )
            try:
                Folder().load(doc["symlinkTargetId"], force=True, exc=True)
            except Exception:
                raise ValidationException(
                    "symlinkTargetId must reference a valid folder",
                    field="symlinkTargetId",
                )

    @access.public(scope=TokenScope.DATA_READ)
    def get_child_folders(self, event):
        params = event.info["params"]
        # Handle Symlink first
        parent_id = params.get("parentId")
        if parent_id and not parent_id.startswith("wtlocal:") and params.get("parentType") == "folder":
            parent = Folder().load(params["parentId"], force=True, exc=True)
            if parent.get("isSymlink") and parent.get("symlinkTargetId"):
                event.info["params"]["parentId"] = str(parent["symlinkTargetId"])
        self._get_child_folders(event)

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def _get_child_folders(self, event, path, root, user=None):
        params = event.info["params"]
        name = params.get("name")
        offset = int(params.get("offset", 0))
        limit = int(params.get("limit", 50))
        sort_key = params.get("sort", "lowerName")
        reverse = int(params.get("sortdir", pymongo.ASCENDING)) == pymongo.DESCENDING

        # TODO: implement "text"
        if name:
            if (path / name).is_dir():
                folders = [self.vFolder(path / name, root)]
            else:
                folders = []
        else:
            folders = [
                self.vFolder(obj, root) for obj in path.iterdir() if obj.is_dir()
            ]

        folders = sorted(folders, key=itemgetter(sort_key), reverse=reverse)
        upper_bound = limit + offset if limit > 0 else None
        response = [
            Folder().filter(folder, user=user) for folder in folders[offset:upper_bound]
        ]
        event.preventDefault().addResponse(response)

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE, validate_admin=True)
    def create_folder(self, event, path, root, user=None):
        params = event.info["params"]
        new_path = path / params["name"]
        try:
            new_path.mkdir()
        except FileExistsError:
            raise ValidationException(
                "A folder with that name already exists here.", "name"
            )
        event.preventDefault().addResponse(
            Folder().filter(self.vFolder(new_path, root), user=user)
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_folder_info(self, event, path, root, user=None):
        event.preventDefault().addResponse(
            Folder().filter(self.vFolder(path, root), user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE, validate_admin=True)
    def rename_folder(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        source = self.vFolder(path, root)

        params = event.info.get("params", {})
        name = params.get("name", path.name)
        parentId = params.get("parentId", source["parentId"])

        if parentId == source["parentId"]:
            new_path = path.with_name(name)
            bail_if_exists(new_path)
            # Just rename in place
            path.rename(new_path)
        else:
            dst_path, dst_root_id = self.path_from_id(parentId)
            # Check whether the user can write to the destination
            Folder().load(dst_root_id, user=user, level=AccessType.WRITE, exc=True)
            new_path = dst_path / name
            bail_if_exists(new_path)
            shutil.move(
                path.as_posix(), new_path.as_posix(), copy_function=shutil.copytree
            )

        event.preventDefault().addResponse(
            Folder().filter(self.vFolder(new_path, root), user=user)
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def remove_folder(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        shutil.rmtree(path.as_posix())
        event.preventDefault().addResponse(
            {"message": "Deleted folder %s." % path.name}
        )

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def remove_folder_contents(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        for sub_path in path.iterdir():
            if sub_path.is_file():
                sub_path.unlink()
            elif sub_path.is_dir():
                shutil.rmtree(sub_path.as_posix())
        event.preventDefault().addResponse(
            {"message": "Deleted contents of folder %s." % path.name}
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def copy_folder(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        source = self.vFolder(path, root)

        if not str(source["_id"]).startswith("wtlocal:"):
            raise GirderException("Copying mappings is not allowed.")

        params = event.info.get("params", {})
        name = params.get("name", path.name)
        parentId = params.get("parentId", source["parentId"])

        dst_path, dst_root_id = self.path_from_id(parentId)
        dst_root = Folder().load(
            dst_root_id, user=user, level=AccessType.WRITE, exc=True
        )
        if "fsPath" not in dst_root:
            raise GirderException("Folder {} is not a mapping.".format(dst_root["_id"]))

        new_path = ensure_unique_path(dst_path, name)
        shutil.copytree(path.as_posix(), new_path.as_posix())
        event.preventDefault().addResponse(
            Folder().filter(self.vFolder(new_path, dst_root), user=user)
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_folder_details(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        response = {"nFolders": 0, "nItems": 0}
        for obj in path.iterdir():
            if obj.is_dir():
                response["nFolders"] += 1
            elif obj.is_file():
                response["nItems"] += 1
        event.preventDefault().addResponse(response)

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def download_folder(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        setResponseHeader("Content-Type", "application/zip")
        setContentDisposition(path.name + ".zip")

        def stream():
            def recursive_file_list(p):
                for obj in p.iterdir():
                    if obj.is_file():
                        yield obj
                    elif obj.is_dir():
                        yield from recursive_file_list(obj)

            zip_stream = ziputil.ZipGenerator(rootPath="")
            for obj in recursive_file_list(path):
                zip_path = os.path.relpath(obj.as_posix(), path.as_posix())
                for data in zip_stream.addFile(lambda: file_stream(obj), zip_path):  # noqa
                    yield data
            yield zip_stream.footer()

        event.preventDefault().addResponse(stream)

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def folder_root_path(self, event, path, root, user=None):
        root_path = pathlib.Path(root["fsPath"])
        response = []
        if root_path != path:
            response.append(
                {
                    "type": "folder",
                    "object": Folder().filter(self.vFolder(path, root), user=user),
                }
            )
            path = path.parent
            while path != root_path:
                response.append(
                    {
                        "type": "folder",
                        "object": Folder().filter(self.vFolder(path, root), user=user),
                    }
                )
                path = path.parent

        response.append({"type": "folder", "object": Folder().filter(root, user=user)})
        girder_rootpath = Folder().parentsToRoot(root, user=self.getCurrentUser())
        response += girder_rootpath[::-1]
        response.pop(0)
        event.preventDefault().addResponse(response[::-1])

    @access.user(scope=TokenScope.DATA_WRITE)
    @validate_event(level=AccessType.WRITE)
    def create_folder_recursive(self, event, path, root, user=None):
        self.is_dir(path, root["_id"])
        new_folder = path / event.info["params"]["path"]
        new_folder.mkdir(parents=True, exist_ok=True)
        event.preventDefault().addResponse(
            Folder().filter(self.vFolder(new_folder, root), user=user)
        )

    @access.public(scope=TokenScope.DATA_READ)
    @validate_event(level=AccessType.READ)
    def get_folder_readme(self, event, path, root, user=None):
        return ""
